//+--------------------------------------------------------------------------------------+
//|                                                       BPNN Matrix Predictor Demo.mq5 |
//|                                             Copyright (c) 2009-2023, gpwr, Marketeer |
//|                                              https://www.mql5.com/en/users/marketeer |
//|                                              https://www.mql5.com/en/users/gpwr      |
//+--------------------------------------------------------------------------------------+
#property copyright "Copyright (c) 2009-2023, gpwr, Marketeer"
#property version "3.0"
#property link "https://www.mql5.com/en/users/marketeer"
#property description "This is not a real world indicator, but a simple demo of BPNN library, originally written in C++ and ported to MQL.\n"
#property description "The demo shows training and testing a neural network for timeseries prediction."


#property indicator_chart_window

#property indicator_buffers 3
#property indicator_plots 3
#property indicator_color1 Red
#property indicator_width1 2
#property indicator_color2 Blue
#property indicator_width2 2
#property indicator_color3 Lime
#property indicator_width3 2


#define BATCH_PROP
#include <MatrixNet.mqh>

enum AF_TYPES
{
  Sigmoid = AF_SIGMOID,
  Tanh = AF_TANH,
  SoftSign = AF_SOFTSIGN,
};
//===================================== INPUTS ===========================================

input int _lastBar = 0;     // Last bar in the past data
input int _futBars = 10;    // # of future bars to predict
input int _smoothPer = 6;   // Smoothing period
input int _numLayers = 3;   // # of layers including input, hidden & output (2..6)
input int _numInputs = 12;  // # of inputs (that is neurons in input 0-th layer)
input int _numNeurons1 = 5; // # of neurons in the 1-st hidden or output layer
input int _numNeurons2 = 1; // # of neurons in the 2-nd hidden or output layer
input int _numNeurons3 = 0; // # of neurons in the 3-rd hidden or output layer
input int _numNeurons4 = 0; // # of neurons in the 4-th hidden or output layer
input int _numNeurons5 = 0; // # of neurons in the 5-th hidden or output layer
input int _ntr = 500;       // # of training sets / bars
input int _nep = 1000;      // Max # of epochs
input int _maxMSEpwr = -7;  // Error (as power of 10) for training to stop; mse < 10^this
input AF_TYPES _AFT = SoftSign; // Activ. function (sigmoid, tanh, x/(1+|x|))
#ifdef BATCH_PROP
input double SpeedUp = 1.1;
input double SpeedDown = 0.1;
input double SpeedHigh = 50;
input double SpeedLow = 0;
#endif


//=================== Helper shorthand methods for indicator setup =======================

bool SetIndexSeries(const int index, double &buffer[], const ENUM_INDEXBUFFER_TYPE type = INDICATOR_DATA)
{
  bool b = ::SetIndexBuffer(index, buffer, type);
  ArraySetAsSeries(buffer, true);
  return b;
}

void SetIndexStyle(const int index, const int type, const int style = -1, const int width = -1, const color clr = clrNONE)
{
  PlotIndexSetInteger(index, PLOT_DRAW_TYPE, type);
  if(style != -1) PlotIndexSetInteger(index, PLOT_LINE_STYLE, style);
  if(width != -1) PlotIndexSetInteger(index, PLOT_LINE_WIDTH, width);
  if(clr != clrNONE) PlotIndexSetInteger(index, PLOT_LINE_COLOR, clr);
}


//======================================= INIT ===========================================
// Indicator buffers
double _pred[], _trainedOut[], _targetOut[];
int _handle;

// Global variables
int _lb, _nf, _nin, _nout, _lSz[], _prevBars, _fdMax;
double _maxMSE;

double _extInitWt[]; // trained weights


int OnInit()
{
    // Create 1D array describing NN -----------------------------------------------------+
    ArrayResize(_lSz, _numLayers);
    _lSz[0] = _numInputs;
    _lSz[1] = _numNeurons1;
    if(_numLayers > 2)
    {
        _lSz[2] = _numNeurons2;
        if(_numLayers > 3)
        {
            _lSz[3] = _numNeurons3;
            if(_numLayers > 4)
            {
                _lSz[4] = _numNeurons4;
                if(_numLayers > 5) _lSz[5] = _numNeurons5;
            }
        }
    }

    // Use shorter names for some external inputs -------------------------------------------+
    _lb = _lastBar;
    _nf = _futBars;
    _nin = _numInputs;
    _nout = _lSz[_numLayers - 1];
    _maxMSE = MathPow(10.0, _maxMSEpwr);
    _prevBars = iBars(_Symbol, _Period) - 1;

    // Find maximum Fibonacci delay ---------------------------------------------------------+
    int fd2 = 0;
    int fd1 = 1;
    for(int j = 0; j < _nin; j++)
    {
        int fd = fd1 + fd2;
        fd2 = fd1;
        fd1 = fd;
    }
    _fdMax = fd1;

    // Set indicator properties -------------------------------------------------------------+
    SetIndexSeries(0, _pred);
    SetIndexStyle(0, DRAW_LINE, STYLE_SOLID, 2);
    PlotIndexSetString(0, PLOT_LABEL, "Prediction");
    SetIndexSeries(1, _trainedOut);
    SetIndexStyle(1, DRAW_LINE, STYLE_SOLID, 2);
    PlotIndexSetString(1, PLOT_LABEL, "Output");
    SetIndexSeries(2, _targetOut);
    SetIndexStyle(2, DRAW_LINE, STYLE_SOLID, 2);
    PlotIndexSetString(2, PLOT_LABEL, "Target");
    PlotIndexSetInteger(0, PLOT_SHIFT, _nf - _lb); // future data vector i=0.._nf; _nf corresponds to bar=_lb
    IndicatorSetString(INDICATOR_SHORTNAME, "BPNN");
    
    _handle = iMA(NULL, 0, _smoothPer, 0, MODE_EMA, PRICE_MEDIAN);
    return (_handle != INVALID_HANDLE) ? INIT_SUCCEEDED : INIT_FAILED;
}

//===================================== DEINIT ===========================================
void OnDeinit(const int)
{
  Comment("");
}

//===================================== START ============================================
int OnCalculate(const int rates_total,
                 const int prev_calculated,
                 const int begin,
                 const double& price[])
{
    int i, j, k;
    int fd, fd1, fd2;

    if(_prevBars < iBars(_Symbol, _Period))
    {
        const int n = _ntr + _fdMax + 1;
        if(n + _lb > iBars(_Symbol, _Period))
        {
           Comment(StringFormat("Insufficient data: %d bars, required: %d+%d", iBars(_Symbol, _Period), n, _lb));
           return 0;
        }
        
        ArrayInitialize(_pred, EMPTY_VALUE);
        ArrayInitialize(_trainedOut, EMPTY_VALUE);
        ArrayInitialize(_targetOut, EMPTY_VALUE);
        
        _prevBars = iBars(_Symbol, _Period);
        // Check NN and find the total number of weights ----------------------------------------+
        if(_numLayers > 6)
        {
            Print("The maximum number of layers is 6");
            return 0;
        }
        for(i = 0; i < _numLayers; i++)
        {
            if(_lSz[i] <= 0)
            {
                Print("No neurons in layer # " + DoubleToString(i, 0) +
                      ". Either reduce # of layers or add neurons to this layer");
                return 0;
            }
        }
        int nw = 0;                               // total number of weights
        for(i = 1; i < _numLayers; i++)           // for each layer except input
            for(j = 0; j < _lSz[i]; j++)          // for each neuron in current layer
                for(k = 0; k <= _lSz[i - 1]; k++) // for each input of current neuron including bias
                    nw++;

        double x[];
        ArrayResize(x, n);

        ResetLastError();
        double value[1];

        // First smooth prices
        for(i = 0; i < n; i++)
        {
            if(CopyBuffer(_handle, 0, _lb + i, 1, value) == -1)
            {
                const int err = GetLastError(); // probably ERR_INDICATOR_DATA_NOT_FOUND
                Print("Error: ", err);
                Comment("Error: ", err, " at ", TimeToString(TimeCurrent()));
                _prevBars = 0;
                // let indicator calculate anew
                return 0;
            }
            else
            {
                x[i] = value[0];
            }
        }

        // Prepare input data for training ------------------------------------------------------+
        double inpTrain[], outTarget[];
        ArrayResize(inpTrain, _ntr * _nin);
        ArrayResize(outTarget, _ntr * _nout);

        // The input data is arranged as follows:
        //
        // inpTrain[i*_nin+j]
        //------------------
        //      j= 0..._nin-1
        //            |
        // i=0     <inputs>
        // ...     <inputs>
        // i=_ntr-1 <inputs>
        //
        // outTarget[i*_nout+j]
        //--------------------
        //      j= 0..._nout-1
        //             |
        // i=0     <targets>
        // ...     <targets>
        // i=_ntr-1 <targets>
        //
        // <inputs> start with the oldest value first
        
        // Fill in the input arrays with data; in this example _nout=1
        for(i = _ntr - 1; i >= 0; i--)
        {
            outTarget[i] = (x[_ntr - 1 - i] / x[_ntr - i] - 1.0);
            fd2 = 0;
            fd1 = 1;
            for(j = _nin - 1; j >= 0; j--)
            {
                fd = fd1 + fd2; // use Fibonacci delays: 1,2,3,5,8,13,21,34,55,89,144...
                fd2 = fd1;
                fd1 = fd;
                inpTrain[i * _nin + j] = x[_ntr - i] / x[_ntr - i + fd] - 1.0;
            }
        }

        // we will re-train on every new bar;
        // in the tester we can use previously obtained weights -
        // this will eliminate lengthy on-the-fly operations, such as neural network training
        const bool alreadyTrained = ArraySize(_extInitWt) == nw && MQLInfoInteger(MQL_TESTER);
        
        if(!alreadyTrained)
        {

        // Train NN -----------------------------------------------------------------------------+
        double outTrain[], trainedWt[];

        // The output data is arranged as follows:
        //
        // outTrain[i*_nout+j]
        //      j= 0..._nout-1
        //             |
        // i=0     <outputs>
        // ...     <outputs>
        // i=_ntr-1 <outputs>

        const bool status = Train(inpTrain, outTarget, outTrain, _ntr, 0, _extInitWt, trainedWt, _numLayers,
                                  _lSz, _AFT, 0, _nep, _maxMSE);
        if(status) // ok
        {
            // Store trainedWt[] as _extInitWt[] for next training
            ArrayCopy(_extInitWt, trainedWt);
        }
        else // if failed, don't save the weights
        {
            ArrayResize(_extInitWt, 0);
        }

        // Show how individual net outputs match targets
        for(i = 0; i < _ntr; i++)
        {
            _trainedOut[_lb + i] = (1.0 + outTrain[_ntr - 1 - i]) * x[i + 1];
        }
        } // !alreadyTrained

        for(i = 0; i < _ntr; i++)
        {
            _targetOut[_lb + i] = x[i];
        }

        // Test NN ------------------------------------------------------------------------------+
        double inpTest[], outTest[];

        // The input data is arranged as follows:
        //
        // inpTest[i*_nin+j]
        //-----------------
        //      j= 0..._nin-1
        //            |
        // i=0     <inputs>
        // ...     <inputs>
        // i=ntt-1 <inputs>
        //
        // <inputs> start with the oldest value first
        //
        // The output data is arranged as follows:
        //
        // outTest[i*_nout+j]
        //------------------
        //      j= 0..._nout-1
        //             |
        // i=0     <outputs>
        // ...     <outputs>
        // i=ntt-1 <outputs>

        _pred[_nf] = x[0];
        for(i = 0; i < _nf; i++)
        {
            ArrayResize(inpTest, _nin);
            fd2 = 0;
            fd1 = 1;
            for(j = _nin - 1; j >= 0; j--)
            {
                fd = fd1 + fd2; // use Fibonacci delays: 1,2,3,5,8,13,21,34,55,89,144...
                fd2 = fd1;
                fd1 = fd;
                double o, od;
                if(i > 0)
                    o = _pred[_nf - i];
                else
                    o = x[0];
                if(i - fd > 0)
                    od = _pred[_nf - i + fd];
                else
                    od = x[fd - i];
                inpTest[j] = o / od - 1.0;
            }
            
            Test(inpTest, outTest, 1, _extInitWt, _numLayers, _lSz, _AFT, 0);
            _pred[_nf - i - 1] = _pred[_nf - i] * (outTest[0] + 1.0); // predicted next open
        }
    }
    return rates_total;
}

bool Train(
  double       &inpTrain[],  // Input training data (2D data as 1D array, oldest first)
  double       &outTarget[], // Output target data for training (2D data as 1D array, oldest first)
  double       &outTrain[],  // Output 1D array to hold net outputs from training
  const int     ntr,         // # of training sets
  const bool    UEW,         // Use External Weights for initialization (1=use extInitWt, 0=use rnd)
  const double &extInitWt[], // Input 1D array to hold 3D array of external initial weights
  double       &trainedWt[], // Output 1D array to hold 3D array of trained weights
  const int     numLayers,   // # of net layers including input, hidden and output
  const int    &lSz[],       // # of neurons in layers. lSz[0] is # of net inputs (nin)
  const int     AFT,         // Type of neuron activation function (0:sigm, 1:tanh, 2:x/(1+x))
  const bool    OAF,         // 1 enables activation function for output layer neurons; 0 disables
  const int     nep,         // Max # of training epochs
  const double  maxMSE       // Max MSE; training stops once maxMSE is reached
)
{
  // NB: UEW and extInitWt are not used here
  
  if(ArraySize(lSz) != numLayers) return false;
  if(ArraySize(lSz) < 2) return false;

  const int inputs = lSz[0];
  const int outputs = lSz[numLayers - 1];
  const ENUM_ACTIVATION_FUNCTION af = (ENUM_ACTIVATION_FUNCTION)AFT;

  MatrixNet net(lSz, af, OAF ? af : AF_NONE);
  
  matrix data, target, output;
  double mse;

  data.Swap(inpTrain);
  data.Reshape(ntr, inputs);
  target.Swap(outTarget);
  target.Reshape(ntr, outputs);
  
  #ifdef BATCH_PROP
  net.setupSpeedAdjustment(SpeedUp, SpeedDown, SpeedHigh, SpeedLow);
  #endif

  mse = net.train(data, target, nep, maxMSE);
  if(!MathIsValidNumber(mse)) return false;
  
  matrix weights[];
  net.getBestWeights(weights);
  MatrixNet best(weights, af, OAF ? af : AF_NONE);
  
  if(!best.feedForward(data)) return false;
  
  output = best.getResults();
  output.Swap(outTrain);
  
  for(int i = 0; i < ArraySize(weights); ++i) // must be (numLayers - 1)
  {
    double temp[];
    weights[i].Swap(temp);
    ArrayCopy(trainedWt, temp, ArraySize(trainedWt));
  }
  return true;
}

bool Test(
    double       &inpTest[],   // Input test data (2D data as 1D array, oldest first)
    double       &outTest[],   // Output 1D array to hold net outputs from training (oldest first)
    const int     ntt,         // # of test sets
    const double &extInitWt[], // Input 1D array to hold 3D array of external initial weights
    const int     numLayers,   // # of layers including input, hidden and output
    const int    &lSz[],       // # of neurons in layers. lSz[0] is # of net inputs
    const int    AFT,          // Type of neuron activation function (0:sigm, 1:tanh, 2:x/(1+x))
    const int    OAF           // 1 enables activation function for output layer; 0 disables
)
{
  // NB: ntt is not checked here
  
  if(ArraySize(lSz) != numLayers) return false;
  if(ArraySize(lSz) < 2) return false;

  const int inputs = lSz[0];
  const int outputs = lSz[numLayers - 1];
  const ENUM_ACTIVATION_FUNCTION af = (ENUM_ACTIVATION_FUNCTION)AFT;
  
  matrix data, output;
  
  data.Swap(inpTest);
  data.Reshape(ntt, inputs);

  int offset = 0;
  matrix weights[];
  ArrayResize(weights, numLayers - 1);
  for(int i = 1; i < numLayers; ++i)
  {
    double temp[];
    offset += ArrayCopy(temp, extInitWt, 0, offset, (lSz[i - 1] + 1) * lSz[i]);
    weights[i - 1].Swap(temp);
    weights[i - 1].Reshape(lSz[i - 1] + 1, lSz[i]);
  }
  MatrixNet best(weights, af, OAF ? af : AF_NONE);

  if(!best.feedForward(data)) return false;
  
  output = best.getResults();
  output.Swap(outTest);

  return true;
}
